#!/usr/bin/env python3
"""
Crypto Vault
Advanced cryptographic toolkit for secure data handling.
Parameters:
    - mode: Operation mode (encrypt, decrypt, analyze)
    - algorithm: Cryptographic algorithm to use
Requires: Strong key management practices
"""
import os
import logging
from datetime import datetime
import json
from pathlib import Path
import threading
import queue
import time
import base64
import hashlib
import hmac
import secrets
from typing import Dict, List, Union, Optional

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.asymmetric import rsa, padding as asym_padding
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
    Encoding,
    PrivateFormat,
    PublicFormat,
    NoEncryption
)

class CryptoVault:
    def __init__(self, mode: str = 'encrypt', algorithm: str = 'AES-256-GCM'):
        self.mode = mode
        self.algorithm = algorithm
        
        self.key_store = {}
        self.operation_queue = queue.Queue()
        self.result_queue = queue.Queue()
        self.stop_event = threading.Event()
        
        # Initialize components
        self.setup_logging()
        self.initialize_keys()
        
    def setup_logging(self):
        """Set up logging configuration"""
        log_dir = Path("crypto_logs")
        log_dir.mkdir(exist_ok=True)
        
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(log_dir / "crypto.log"),
                logging.StreamHandler()
            ]
        )
        
    def initialize_keys(self):
        """Initialize cryptographic keys"""
        keys_dir = Path("crypto_keys")
        keys_dir.mkdir(exist_ok=True)
        
        # Generate master key if not exists
        master_key_file = keys_dir / "master.key"
        if not master_key_file.exists():
            master_key = Fernet.generate_key()
            with open(master_key_file, 'wb') as f:
                f.write(master_key)
                
        # Load master key
        with open(master_key_file, 'rb') as f:
            self.master_key = f.read()
            
        # Initialize Fernet for key encryption
        self.key_fernet = Fernet(self.master_key)
        
        # Generate RSA key pair if not exists
        rsa_private_file = keys_dir / "rsa_private.pem"
        rsa_public_file = keys_dir / "rsa_public.pem"
        
        if not rsa_private_file.exists() or not rsa_public_file.exists():
            private_key = rsa.generate_private_key(
                public_exponent=65537,
                key_size=2048
            )
            
            public_key = private_key.public_key()
            
            # Save private key
            with open(rsa_private_file, 'wb') as f:
                f.write(private_key.private_bytes(
                    encoding=Encoding.PEM,
                    format=PrivateFormat.PKCS8,
                    encryption_algorithm=NoEncryption()
                ))
                
            # Save public key
            with open(rsa_public_file, 'wb') as f:
                f.write(public_key.public_bytes(
                    encoding=Encoding.PEM,
                    format=PublicFormat.SubjectPublicKeyInfo
                ))
                
    def generate_key(self, size: int = 32) -> bytes:
        """Generate random key of specified size"""
        return secrets.token_bytes(size)
        
    def derive_key(self, password: str, salt: Optional[bytes] = None) -> tuple:
        """Derive key from password using PBKDF2"""
        if salt is None:
            salt = os.urandom(16)
            
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=100000
        )
        
        key = kdf.derive(password.encode())
        return key, salt
        
    def encrypt_key(self, key: bytes) -> bytes:
        """Encrypt key using master key"""
        return self.key_fernet.encrypt(key)
        
    def decrypt_key(self, encrypted_key: bytes) -> bytes:
        """Decrypt key using master key"""
        return self.key_fernet.decrypt(encrypted_key)
        
    def store_key(self, key_id: str, key: bytes):
        """Store encrypted key"""
        encrypted_key = self.encrypt_key(key)
        self.key_store[key_id] = encrypted_key
        
    def retrieve_key(self, key_id: str) -> Optional[bytes]:
        """Retrieve and decrypt key"""
        if key_id in self.key_store:
            encrypted_key = self.key_store[key_id]
            return self.decrypt_key(encrypted_key)
        return None
        
    def aes_encrypt(self, data: bytes, key: Optional[bytes] = None) -> Dict:
        """Encrypt data using AES"""
        if key is None:
            key = self.generate_key()
            
        iv = os.urandom(12)  # 96 bits for GCM
        
        cipher = Cipher(
            algorithms.AES(key),
            modes.GCM(iv)
        )
        
        encryptor = cipher.encryptor()
        ciphertext = encryptor.update(data) + encryptor.finalize()
        
        return {
            'algorithm': 'AES-256-GCM',
            'key': key,
            'iv': iv,
            'tag': encryptor.tag,
            'ciphertext': ciphertext
        }
        
    def aes_decrypt(self, encrypted_data: Dict) -> bytes:
        """Decrypt data using AES"""
        cipher = Cipher(
            algorithms.AES(encrypted_data['key']),
            modes.GCM(encrypted_data['iv'], encrypted_data['tag'])
        )
        
        decryptor = cipher.decryptor()
        return decryptor.update(encrypted_data['ciphertext']) + decryptor.finalize()
        
    def rsa_encrypt(self, data: bytes) -> Dict:
        """Encrypt data using RSA"""
        # Load public key
        with open(Path("crypto_keys") / "rsa_public.pem", 'rb') as f:
            public_key = load_pem_public_key(f.read())
            
        # Generate session key
        session_key = self.generate_key()
        
        # Encrypt session key with RSA
        encrypted_key = public_key.encrypt(
            session_key,
            asym_padding.OAEP(
                mgf=asym_padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        
        # Encrypt data with session key
        encrypted_data = self.aes_encrypt(data, session_key)
        
        return {
            'algorithm': 'RSA-2048',
            'encrypted_key': encrypted_key,
            'iv': encrypted_data['iv'],
            'tag': encrypted_data['tag'],
            'ciphertext': encrypted_data['ciphertext']
        }
        
    def rsa_decrypt(self, encrypted_data: Dict) -> bytes:
        """Decrypt data using RSA"""
        # Load private key
        with open(Path("crypto_keys") / "rsa_private.pem", 'rb') as f:
            private_key = load_pem_private_key(f.read(), password=None)
            
        # Decrypt session key
        session_key = private_key.decrypt(
            encrypted_data['encrypted_key'],
            asym_padding.OAEP(
                mgf=asym_padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        
        # Decrypt data with session key
        return self.aes_decrypt({
            'key': session_key,
            'iv': encrypted_data['iv'],
            'tag': encrypted_data['tag'],
            'ciphertext': encrypted_data['ciphertext']
        })
        
    def hash_data(self, data: bytes, algorithm: str = 'SHA-256') -> Dict:
        """Hash data using specified algorithm"""
        algorithms = {
            'SHA-256': hashlib.sha256,
            'SHA-384': hashlib.sha384,
            'SHA-512': hashlib.sha512,
            'BLAKE2b': hashlib.blake2b,
            'BLAKE2s': hashlib.blake2s
        }
        
        if algorithm not in algorithms:
            raise ValueError(f"Unsupported hash algorithm: {algorithm}")
            
        hash_func = algorithms[algorithm]()
        hash_func.update(data)
        
        return {
            'algorithm': algorithm,
            'hash': hash_func.hexdigest()
        }
        
    def sign_data(self, data: bytes) -> Dict:
        """Sign data using RSA"""
        # Load private key
        with open(Path("crypto_keys") / "rsa_private.pem", 'rb') as f:
            private_key = load_pem_private_key(f.read(), password=None)
            
        # Create signature
        signature = private_key.sign(
            data,
            asym_padding.PSS(
                mgf=asym_padding.MGF1(hashes.SHA256()),
                salt_length=asym_padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )
        
        return {
            'algorithm': 'RSA-PSS',
            'signature': signature
        }
        
    def verify_signature(self, data: bytes, signature_data: Dict) -> bool:
        """Verify RSA signature"""
        # Load public key
        with open(Path("crypto_keys") / "rsa_public.pem", 'rb') as f:
            public_key = load_pem_public_key(f.read())
            
        try:
            public_key.verify(
                signature_data['signature'],
                data,
                asym_padding.PSS(
                    mgf=asym_padding.MGF1(hashes.SHA256()),
                    salt_length=asym_padding.PSS.MAX_LENGTH
                ),
                hashes.SHA256()
            )
            return True
        except:
            return False
            
    def analyze_encryption(self, encrypted_data: Dict) -> Dict:
        """Analyze encrypted data"""
        analysis = {
            'algorithm': encrypted_data.get('algorithm', 'unknown'),
            'key_size': None,
            'iv_size': None,
            'tag_size': None,
            'ciphertext_size': None,
            'total_size': 0
        }
        
        # Analyze components
        if 'key' in encrypted_data:
            analysis['key_size'] = len(encrypted_data['key'])
            analysis['total_size'] += analysis['key_size']
            
        if 'iv' in encrypted_data:
            analysis['iv_size'] = len(encrypted_data['iv'])
            analysis['total_size'] += analysis['iv_size']
            
        if 'tag' in encrypted_data:
            analysis['tag_size'] = len(encrypted_data['tag'])
            analysis['total_size'] += analysis['tag_size']
            
        if 'ciphertext' in encrypted_data:
            analysis['ciphertext_size'] = len(encrypted_data['ciphertext'])
            analysis['total_size'] += analysis['ciphertext_size']
            
        return analysis
        
    def process_operation(self, operation: Dict):
        """Process cryptographic operation"""
        try:
            result = {
                'timestamp': datetime.now().isoformat(),
                'operation': operation['type'],
                'success': False,
                'result': None,
                'error': None
            }
            
            if operation['type'] == 'encrypt':
                if self.algorithm == 'AES-256-GCM':
                    encrypted = self.aes_encrypt(operation['data'])
                else:  # RSA-2048
                    encrypted = self.rsa_encrypt(operation['data'])
                    
                key_id = secrets.token_hex(16)
                self.store_key(key_id, encrypted['key'])
                
                result.update({
                    'success': True,
                    'result': {
                        'key_id': key_id,
                        'encrypted_data': encrypted
                    }
                })
                
            elif operation['type'] == 'decrypt':
                key = self.retrieve_key(operation['key_id'])
                if key:
                    if self.algorithm == 'AES-256-GCM':
                        operation['encrypted_data']['key'] = key
                        decrypted = self.aes_decrypt(operation['encrypted_data'])
                    else:  # RSA-2048
                        decrypted = self.rsa_decrypt(operation['encrypted_data'])
                        
                    result.update({
                        'success': True,
                        'result': {
                            'decrypted_data': decrypted
                        }
                    })
                else:
                    result['error'] = 'Key not found'
                    
            elif operation['type'] == 'hash':
                hashed = self.hash_data(
                    operation['data'],
                    operation.get('algorithm', 'SHA-256')
                )
                result.update({
                    'success': True,
                    'result': hashed
                })
                
            elif operation['type'] == 'sign':
                signed = self.sign_data(operation['data'])
                result.update({
                    'success': True,
                    'result': signed
                })
                
            elif operation['type'] == 'verify':
                verified = self.verify_signature(
                    operation['data'],
                    operation['signature_data']
                )
                result.update({
                    'success': True,
                    'result': {
                        'verified': verified
                    }
                })
                
            elif operation['type'] == 'analyze':
                analysis = self.analyze_encryption(operation['encrypted_data'])
                result.update({
                    'success': True,
                    'result': analysis
                })
                
            self.result_queue.put(result)
            
        except Exception as e:
            logging.error(f"Error processing operation: {str(e)}")
            self.result_queue.put({
                'timestamp': datetime.now().isoformat(),
                'operation': operation['type'],
                'success': False,
                'error': str(e)
            })
            
    def start_processing(self):
        """Start cryptographic processing"""
        logging.info("Starting Crypto Vault")
        logging.info(f"Mode: {self.mode}")
        logging.info(f"Algorithm: {self.algorithm}")
        
        try:
            while not self.stop_event.is_set():
                try:
                    # Process operation
                    operation = self.operation_queue.get(timeout=1)
                    self.process_operation(operation)
                    
                except queue.Empty:
                    continue
                    
        except KeyboardInterrupt:
            logging.info("Stopping Crypto Vault...")
            self.stop_event.set()
            
            # Save final state
            self.save_state()
            
    def save_state(self):
        """Save vault state and results"""
        state = {
            'timestamp': datetime.now().isoformat(),
            'mode': self.mode,
            'algorithm': self.algorithm,
            'keys': {
                kid: base64.b64encode(key).decode()
                for kid, key in self.key_store.items()
            }
        }
        
        # Get all results
        results = []
        while not self.result_queue.empty():
            result = self.result_queue.get()
            
            # Convert bytes to base64 for JSON serialization
            if result.get('result'):
                if isinstance(result['result'], dict):
                    for key, value in result['result'].items():
                        if isinstance(value, bytes):
                            result['result'][key] = base64.b64encode(value).decode()
                            
            results.append(result)
            
        state['results'] = results
        
        # Save to file
        output_dir = Path("crypto_logs")
        output_file = output_dir / f"state_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
        with open(output_file, 'w') as f:
            json.dump(state, f, indent=2)
            
        logging.info(f"State saved to {output_file}")
        
        # Print summary
        print("\nCrypto Vault Summary")
        print("=" * 50)
        print(f"Mode: {self.mode}")
        print(f"Algorithm: {self.algorithm}")
        print(f"Total Operations: {len(results)}")
        
        if results:
            print("\nOperation Summary:")
            op_types = {}
            for result in results:
                op_type = result['operation']
                success = result['success']
                op_types[op_type] = op_types.get(op_type, {'success': 0, 'failure': 0})
                if success:
                    op_types[op_type]['success'] += 1
                else:
                    op_types[op_type]['failure'] += 1
                    
            for op_type, stats in op_types.items():
                print(f"{op_type}:")
                print(f"  Success: {stats['success']}")
                print(f"  Failure: {stats['failure']}")

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description="Crypto Vault")
    parser.add_argument("--mode", choices=['encrypt', 'decrypt', 'analyze'],
                      default='encrypt', help="Operation mode")
    parser.add_argument("--algorithm", choices=['AES-256-GCM', 'RSA-2048'],
                      default='AES-256-GCM', help="Cryptographic algorithm")
    args = parser.parse_args()
    
    vault = CryptoVault(args.mode, args.algorithm)
    vault.start_processing()
